#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <assert.h>
#include "odicon.h"
#include "quadratic.h"

/* L, Lg, M, Mg, and others */
double quadratic(const int dim, int *x, double **a) {
    double ret = 0;
    int i, j;

    for (i=0; i<dim; i++) {
        double tmp = 0;
        for (j=0; j<dim; j++) {
            tmp += a[i][j] * x[j];
        }
        ret += tmp * x[i];
    }
    return ret/2;
}

double quadratic2(const int dim, int *x, double **a, double *b) {
    double ret = 0;
    int i;

    for (i=0; i<dim; i++) ret += x[i] * b[i];
    return quadratic(dim, x, a) + ret;
}

double quadratic3(const int dim, int *x, double **a, double *b, double c) {
    return quadratic2(dim, x, a, b) + c;
}

/* easy book pp.41 (4.21) */
int lgconv_validate_hesse(const int dim, double **a) {
    int i, j;

    for (i=0; i<dim; i++) {
        for (j=0; j<i; j++) {
            if (fcmp(a[i][j],a[j][i]) != 0) return FALSE;
            if (fcmp(a[i][j],0) > 0)        return FALSE;
        }
        double sum = 0;
        for (j=0; j<dim; j++) sum += a[i][j];
        if (fcmp(sum,0) < 0) return FALSE;
    }
    return TRUE;
}

/* easy book pp.42 (4.23) */
int lconv_validate_hesse(const int dim, double **a) {
    int i, j;

    for (i=0; i<dim; i++) {
        for (j=0; j<i; j++) {
            if (fcmp(a[i][j],a[j][i]) != 0) return FALSE;
            if (fcmp(a[i][j],0) > 0)        return FALSE;
        }
        double sum = 0;
        for (j=0; j<dim; j++) sum += a[i][j];
        if (fcmp(sum,0) != 0) return FALSE;
    }
    return TRUE;
}

/* easy book pp.65 (5.15)(5.16) */
int mgconv_validate_hesse(const int dim, double **a) {
    int i, j, k;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            double tmp = a[i][j];
            if (fcmp(tmp,a[j][i]) != 0) return FALSE;
            if (fcmp(tmp,0) < 0)        return FALSE;
            for (k=0; k<dim; k++) {
                if (k != i && k != j &&
                    fcmp(tmp,a[i][k]) < 0 &&
                    fcmp(tmp,a[j][k]) < 0) return FALSE;
            }
        }
    }
    return TRUE;
}

/* easy book pp.68 (5.27) */
int mconv_validate_hesse(const int dim, double **a) {
    int i, j, k, l;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            if (fcmp(a[i][j],a[j][i]) != 0) return FALSE;
    for (k=0; k<dim; k++) {
        if (k==i || k==j) continue;
        for (l=0; l<=k; l++) {
            if (l==i || l==j) continue;
            if (fcmp(a[i][j]+a[k][l],a[i][k]+a[j][l]) < 0 &&
                fcmp(a[i][j]+a[k][l],a[i][l]+a[j][k]) < 0) return FALSE;
        }
    }                    
        }
    }                    
    return TRUE;
}

/* L, Lg */
/* easy book pp.42 (4.26) */
void make_quadratic_hesse_l(const int dim, double f(int,int*),
                            double **a) {
    int i, j;
    int *x = ivector_new(dim);

    for (i=0; i<dim; i++) x[i] = 0;

    for (i=0; i<dim; i++) {
        for (j=0; j<i; j++) {
            double tmp = 0;
            tmp -= f(dim,x);  x[i]++;
            tmp += f(dim,x);  x[j]++;
            tmp -= f(dim,x);  x[i]--;
            tmp += f(dim,x);  x[j]--;
            a[i][j] = a[j][i] = -tmp;
        }
    }
    /* diagonal elements */
    for (i=0; i<dim; i++) {
        double tmp = 0;
        tmp += f(dim,x);  x[i]++;
        tmp -= f(dim,x);  for (j=0; j<dim; j++) x[j]++;
        tmp += f(dim,x);  x[i]--;
        tmp -= f(dim,x);  for (j=0; j<dim; j++) x[j]--;
        for (j=0; j<dim; j++) if (i != j) tmp -= a[i][j];
        a[i][i] = tmp;
    }
    free(x);
}


/* M, Mg */
/* easy book pp.69 (5.28) */
void make_quadratic_hesse_mg(const int dim, double f(int,int*),
                             double **a) {
    int i, j;
    int *x = ivector_new(dim);

    for (i=0; i<dim; i++) x[i] = 0;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            double tmp = 0;
            tmp -= f(dim,x);  x[i]++;
            tmp += f(dim,x);  x[j]++;
            tmp -= f(dim,x);  x[i]--;
            tmp += f(dim,x);  x[j]--;
            a[i][j] = a[j][i] = -tmp;
        }
    }
    free(x);
}


/* M */
/* Tree Metric */
void make_quadratic_hesse_m(const int dim, double f(int,int*),
                            double **a) {
    int i, j;
    int *x = ivector_new(dim);

    for (i=0; i<dim; i++) x[i] = 0;

    for (i=0; i<dim; i++) {
        for (j=0; j<=i; j++) {
            double tmp = 0;
            x[i]+=2;          tmp -= f(dim,x)/2;
            x[i]--;  x[j]++;  tmp += f(dim,x);
            x[i]--;  x[j]++;  tmp -= f(dim,x)/2;
                     x[j]-=2;
            a[i][j] = a[j][i] = tmp;
            /*
            tmp -= f(dim,x);  x[i]++;
            tmp += f(dim,x);  x[j]++;
            tmp -= f(dim,x);  x[i]--;
            tmp += f(dim,x);  x[j]--;
            a[i][j] = a[j][i] = tmp;
            */
        }
    }
    free(x);
}


void make_quadratic_liner_const(const int dim, double f(int,int*),
                                double **a, double *b, double *c) {
    int i, step = 10;
    double tmp;
    int *x = ivector_new(dim);

    for (i=0; i<dim; i++) x[i] = 0;

    tmp = f(dim,x) - quadratic(dim,x,a);
    for (i=0; i<dim; i++) {
        x[i] += step;
        b[i] = (f(dim,x) - quadratic(dim,x,a) - tmp) / step;
        x[i] -= step;
    }

    *c = f(dim,x) - quadratic2(dim,x,a,b);
    free(x);
}
